/****************************************************************************
** Meta object code from reading C++ file 'CMenu.h'
**
** Created by: The Qt Meta Object Compiler version 63 (Qt 4.8.7)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "CMenu.h"
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'CMenu.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 63
#error "This file was generated using the moc from 4.8.7. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

QT_BEGIN_MOC_NAMESPACE
static const uint qt_meta_data_MyAction[] = {

 // content:
       6,       // revision
       0,       // classname
       0,    0, // classinfo
       0,    0, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

       0        // eod
};

static const char qt_meta_stringdata_MyAction[] = {
    "MyAction\0"
};

void MyAction::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    Q_UNUSED(_o);
    Q_UNUSED(_id);
    Q_UNUSED(_c);
    Q_UNUSED(_a);
}

const QMetaObjectExtraData MyAction::staticMetaObjectExtraData = {
    0,  qt_static_metacall 
};

const QMetaObject MyAction::staticMetaObject = {
    { &QAction::staticMetaObject, qt_meta_stringdata_MyAction,
      qt_meta_data_MyAction, &staticMetaObjectExtraData }
};

#ifdef Q_NO_DATA_RELOCATION
const QMetaObject &MyAction::getStaticMetaObject() { return staticMetaObject; }
#endif //Q_NO_DATA_RELOCATION

const QMetaObject *MyAction::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->metaObject : &staticMetaObject;
}

void *MyAction::qt_metacast(const char *_clname)
{
    if (!_clname) return 0;
    if (!strcmp(_clname, qt_meta_stringdata_MyAction))
        return static_cast<void*>(const_cast< MyAction*>(this));
    return QAction::qt_metacast(_clname);
}

int MyAction::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QAction::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    return _id;
}
static const uint qt_meta_data_CMenu[] = {

 // content:
       6,       // revision
       0,       // classname
       0,    0, // classinfo
       5,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

 // slots: signature, parameters, type, tag, flags
       7,    6,    6,    6, 0x0a,
      23,    6,    6,    6, 0x0a,
      41,    6,    6,    6, 0x0a,
      57,    6,    6,    6, 0x0a,
      69,    6,    6,    6, 0x0a,

       0        // eod
};

static const char qt_meta_stringdata_CMenu[] = {
    "CMenu\0\0slotTriggered()\0slotToggled(bool)\0"
    "slotDestroyed()\0slotShown()\0slotHidden()\0"
};

void CMenu::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        Q_ASSERT(staticMetaObject.cast(_o));
        CMenu *_t = static_cast<CMenu *>(_o);
        switch (_id) {
        case 0: _t->slotTriggered(); break;
        case 1: _t->slotToggled((*reinterpret_cast< bool(*)>(_a[1]))); break;
        case 2: _t->slotDestroyed(); break;
        case 3: _t->slotShown(); break;
        case 4: _t->slotHidden(); break;
        default: ;
        }
    }
}

const QMetaObjectExtraData CMenu::staticMetaObjectExtraData = {
    0,  qt_static_metacall 
};

const QMetaObject CMenu::staticMetaObject = {
    { &QObject::staticMetaObject, qt_meta_stringdata_CMenu,
      qt_meta_data_CMenu, &staticMetaObjectExtraData }
};

#ifdef Q_NO_DATA_RELOCATION
const QMetaObject &CMenu::getStaticMetaObject() { return staticMetaObject; }
#endif //Q_NO_DATA_RELOCATION

const QMetaObject *CMenu::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->metaObject : &staticMetaObject;
}

void *CMenu::qt_metacast(const char *_clname)
{
    if (!_clname) return 0;
    if (!strcmp(_clname, qt_meta_stringdata_CMenu))
        return static_cast<void*>(const_cast< CMenu*>(this));
    return QObject::qt_metacast(_clname);
}

int CMenu::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 5)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 5;
    }
    return _id;
}
QT_END_MOC_NAMESPACE
