/****************************************************************************
** Meta object code from reading C++ file 'cwebview.h'
**
** Created by: The Qt Meta Object Compiler version 63 (Qt 4.8.7)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "cwebview.h"
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'cwebview.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 63
#error "This file was generated using the moc from 4.8.7. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

QT_BEGIN_MOC_NAMESPACE
static const uint qt_meta_data_MyWebPage[] = {

 // content:
       6,       // revision
       0,       // classname
       0,    0, // classinfo
       0,    0, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

       0        // eod
};

static const char qt_meta_stringdata_MyWebPage[] = {
    "MyWebPage\0"
};

void MyWebPage::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    Q_UNUSED(_o);
    Q_UNUSED(_id);
    Q_UNUSED(_c);
    Q_UNUSED(_a);
}

const QMetaObjectExtraData MyWebPage::staticMetaObjectExtraData = {
    0,  qt_static_metacall 
};

const QMetaObject MyWebPage::staticMetaObject = {
    { &QWebPage::staticMetaObject, qt_meta_stringdata_MyWebPage,
      qt_meta_data_MyWebPage, &staticMetaObjectExtraData }
};

#ifdef Q_NO_DATA_RELOCATION
const QMetaObject &MyWebPage::getStaticMetaObject() { return staticMetaObject; }
#endif //Q_NO_DATA_RELOCATION

const QMetaObject *MyWebPage::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->metaObject : &staticMetaObject;
}

void *MyWebPage::qt_metacast(const char *_clname)
{
    if (!_clname) return 0;
    if (!strcmp(_clname, qt_meta_stringdata_MyWebPage))
        return static_cast<void*>(const_cast< MyWebPage*>(this));
    return QWebPage::qt_metacast(_clname);
}

int MyWebPage::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QWebPage::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    return _id;
}
static const uint qt_meta_data_MyWebView[] = {

 // content:
       6,       // revision
       0,       // classname
       0,    0, // classinfo
       0,    0, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

       0        // eod
};

static const char qt_meta_stringdata_MyWebView[] = {
    "MyWebView\0"
};

void MyWebView::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    Q_UNUSED(_o);
    Q_UNUSED(_id);
    Q_UNUSED(_c);
    Q_UNUSED(_a);
}

const QMetaObjectExtraData MyWebView::staticMetaObjectExtraData = {
    0,  qt_static_metacall 
};

const QMetaObject MyWebView::staticMetaObject = {
    { &QWebView::staticMetaObject, qt_meta_stringdata_MyWebView,
      qt_meta_data_MyWebView, &staticMetaObjectExtraData }
};

#ifdef Q_NO_DATA_RELOCATION
const QMetaObject &MyWebView::getStaticMetaObject() { return staticMetaObject; }
#endif //Q_NO_DATA_RELOCATION

const QMetaObject *MyWebView::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->metaObject : &staticMetaObject;
}

void *MyWebView::qt_metacast(const char *_clname)
{
    if (!_clname) return 0;
    if (!strcmp(_clname, qt_meta_stringdata_MyWebView))
        return static_cast<void*>(const_cast< MyWebView*>(this));
    return QWebView::qt_metacast(_clname);
}

int MyWebView::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QWebView::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    return _id;
}
static const uint qt_meta_data_CWebView[] = {

 // content:
       6,       // revision
       0,       // classname
       0,    0, // classinfo
      13,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

 // slots: signature, parameters, type, tag, flags
      10,    9,    9,    9, 0x0a,
      27,   24,    9,    9, 0x0a,
      55,   46,    9,    9, 0x0a,
      73,    9,    9,    9, 0x0a,
      87,    9,    9,    9, 0x0a,
     111,  106,    9,    9, 0x0a,
     143,  137,    9,    9, 0x0a,
     188,  165,    9,    9, 0x0a,
     225,    9,    9,    9, 0x0a,
     270,  250,    9,    9, 0x0a,
     325,    9,    9,    9, 0x0a,
     342,    9,    9,    9, 0x0a,
     377,    9,    9,    9, 0x0a,

       0        // eod
};

static const char qt_meta_stringdata_CWebView[] = {
    "CWebView\0\0iconChanged()\0ok\0"
    "loadFinished(bool)\0progress\0"
    "loadProgress(int)\0loadStarted()\0"
    "selectionChanged()\0text\0"
    "statusBarMessage(QString)\0title\0"
    "titleChanged(QString)\0link,title,textContent\0"
    "linkHovered(QString,QString,QString)\0"
    "frameCreated(QWebFrame*)\0reply,authenticator\0"
    "authenticationRequired(QNetworkReply*,QAuthenticator*)\0"
    "urlChanged(QUrl)\0downloadRequested(QNetworkRequest)\0"
    "handleUnsupportedContent(QNetworkReply*)\0"
};

void CWebView::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        Q_ASSERT(staticMetaObject.cast(_o));
        CWebView *_t = static_cast<CWebView *>(_o);
        switch (_id) {
        case 0: _t->iconChanged(); break;
        case 1: _t->loadFinished((*reinterpret_cast< bool(*)>(_a[1]))); break;
        case 2: _t->loadProgress((*reinterpret_cast< int(*)>(_a[1]))); break;
        case 3: _t->loadStarted(); break;
        case 4: _t->selectionChanged(); break;
        case 5: _t->statusBarMessage((*reinterpret_cast< const QString(*)>(_a[1]))); break;
        case 6: _t->titleChanged((*reinterpret_cast< const QString(*)>(_a[1]))); break;
        case 7: _t->linkHovered((*reinterpret_cast< const QString(*)>(_a[1])),(*reinterpret_cast< const QString(*)>(_a[2])),(*reinterpret_cast< const QString(*)>(_a[3]))); break;
        case 8: _t->frameCreated((*reinterpret_cast< QWebFrame*(*)>(_a[1]))); break;
        case 9: _t->authenticationRequired((*reinterpret_cast< QNetworkReply*(*)>(_a[1])),(*reinterpret_cast< QAuthenticator*(*)>(_a[2]))); break;
        case 10: _t->urlChanged((*reinterpret_cast< const QUrl(*)>(_a[1]))); break;
        case 11: _t->downloadRequested((*reinterpret_cast< const QNetworkRequest(*)>(_a[1]))); break;
        case 12: _t->handleUnsupportedContent((*reinterpret_cast< QNetworkReply*(*)>(_a[1]))); break;
        default: ;
        }
    }
}

const QMetaObjectExtraData CWebView::staticMetaObjectExtraData = {
    0,  qt_static_metacall 
};

const QMetaObject CWebView::staticMetaObject = {
    { &QObject::staticMetaObject, qt_meta_stringdata_CWebView,
      qt_meta_data_CWebView, &staticMetaObjectExtraData }
};

#ifdef Q_NO_DATA_RELOCATION
const QMetaObject &CWebView::getStaticMetaObject() { return staticMetaObject; }
#endif //Q_NO_DATA_RELOCATION

const QMetaObject *CWebView::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->metaObject : &staticMetaObject;
}

void *CWebView::qt_metacast(const char *_clname)
{
    if (!_clname) return 0;
    if (!strcmp(_clname, qt_meta_stringdata_CWebView))
        return static_cast<void*>(const_cast< CWebView*>(this));
    return QObject::qt_metacast(_clname);
}

int CWebView::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 13)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 13;
    }
    return _id;
}
QT_END_MOC_NAMESPACE
