/****************************************************************************
** Meta object code from reading C++ file 'gview.h'
**
** Created by: The Qt Meta Object Compiler version 63 (Qt 4.8.7)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "gview.h"
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'gview.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 63
#error "This file was generated using the moc from 4.8.7. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

QT_BEGIN_MOC_NAMESPACE
static const uint qt_meta_data_GEditor[] = {

 // content:
       6,       // revision
       0,       // classname
       0,    0, // classinfo
       9,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       4,       // signalCount

 // signals: signature, parameters, type, tag, flags
       9,    8,    8,    8, 0x05,
      23,    8,    8,    8, 0x05,
      37,    8,    8,    8, 0x05,
      56,    8,    8,    8, 0x05,

 // slots: signature, parameters, type, tag, flags
      81,    8,    8,    8, 0x08,
     101,    8,    8,    8, 0x08,
     122,    8,    8,    8, 0x08,
     132,    8,    8,    8, 0x08,
     154,    8,    8,    8, 0x08,

       0        // eod
};

static const char qt_meta_stringdata_GEditor[] = {
    "GEditor\0\0cursorMoved()\0textChanged()\0"
    "marginClicked(int)\0marginDoubleClicked(int)\0"
    "blinkTimerTimeout()\0scrollTimerTimeout()\0"
    "unflash()\0docTextChangedLater()\0"
    "ensureCursorVisible()\0"
};

void GEditor::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        Q_ASSERT(staticMetaObject.cast(_o));
        GEditor *_t = static_cast<GEditor *>(_o);
        switch (_id) {
        case 0: _t->cursorMoved(); break;
        case 1: _t->textChanged(); break;
        case 2: _t->marginClicked((*reinterpret_cast< int(*)>(_a[1]))); break;
        case 3: _t->marginDoubleClicked((*reinterpret_cast< int(*)>(_a[1]))); break;
        case 4: _t->blinkTimerTimeout(); break;
        case 5: _t->scrollTimerTimeout(); break;
        case 6: _t->unflash(); break;
        case 7: _t->docTextChangedLater(); break;
        case 8: _t->ensureCursorVisible(); break;
        default: ;
        }
    }
}

const QMetaObjectExtraData GEditor::staticMetaObjectExtraData = {
    0,  qt_static_metacall 
};

const QMetaObject GEditor::staticMetaObject = {
    { &Q3ScrollView::staticMetaObject, qt_meta_stringdata_GEditor,
      qt_meta_data_GEditor, &staticMetaObjectExtraData }
};

#ifdef Q_NO_DATA_RELOCATION
const QMetaObject &GEditor::getStaticMetaObject() { return staticMetaObject; }
#endif //Q_NO_DATA_RELOCATION

const QMetaObject *GEditor::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->metaObject : &staticMetaObject;
}

void *GEditor::qt_metacast(const char *_clname)
{
    if (!_clname) return 0;
    if (!strcmp(_clname, qt_meta_stringdata_GEditor))
        return static_cast<void*>(const_cast< GEditor*>(this));
    return Q3ScrollView::qt_metacast(_clname);
}

int GEditor::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = Q3ScrollView::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 9)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 9;
    }
    return _id;
}

// SIGNAL 0
void GEditor::cursorMoved()
{
    QMetaObject::activate(this, &staticMetaObject, 0, 0);
}

// SIGNAL 1
void GEditor::textChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, 0);
}

// SIGNAL 2
void GEditor::marginClicked(int _t1)
{
    void *_a[] = { 0, const_cast<void*>(reinterpret_cast<const void*>(&_t1)) };
    QMetaObject::activate(this, &staticMetaObject, 2, _a);
}

// SIGNAL 3
void GEditor::marginDoubleClicked(int _t1)
{
    void *_a[] = { 0, const_cast<void*>(reinterpret_cast<const void*>(&_t1)) };
    QMetaObject::activate(this, &staticMetaObject, 3, _a);
}
QT_END_MOC_NAMESPACE
