/****************************************************************************
** Meta object code from reading C++ file 'cwebview.h'
**
** Created by: The Qt Meta Object Compiler version 67 (Qt 5.9.4)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "cwebview.h"
#include <QtCore/qbytearray.h>
#include <QtCore/qmetatype.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'cwebview.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 67
#error "This file was generated using the moc from 5.9.4. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

QT_BEGIN_MOC_NAMESPACE
QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
struct qt_meta_stringdata_MyWebPage_t {
    QByteArrayData data[1];
    char stringdata0[10];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_MyWebPage_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_MyWebPage_t qt_meta_stringdata_MyWebPage = {
    {
QT_MOC_LITERAL(0, 0, 9) // "MyWebPage"

    },
    "MyWebPage"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_MyWebPage[] = {

 // content:
       7,       // revision
       0,       // classname
       0,    0, // classinfo
       0,    0, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

       0        // eod
};

void MyWebPage::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    Q_UNUSED(_o);
    Q_UNUSED(_id);
    Q_UNUSED(_c);
    Q_UNUSED(_a);
}

const QMetaObject MyWebPage::staticMetaObject = {
    { &QWebPage::staticMetaObject, qt_meta_stringdata_MyWebPage.data,
      qt_meta_data_MyWebPage,  qt_static_metacall, nullptr, nullptr}
};


const QMetaObject *MyWebPage::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *MyWebPage::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_MyWebPage.stringdata0))
        return static_cast<void*>(this);
    return QWebPage::qt_metacast(_clname);
}

int MyWebPage::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QWebPage::qt_metacall(_c, _id, _a);
    return _id;
}
struct qt_meta_stringdata_MyWebView_t {
    QByteArrayData data[1];
    char stringdata0[10];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_MyWebView_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_MyWebView_t qt_meta_stringdata_MyWebView = {
    {
QT_MOC_LITERAL(0, 0, 9) // "MyWebView"

    },
    "MyWebView"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_MyWebView[] = {

 // content:
       7,       // revision
       0,       // classname
       0,    0, // classinfo
       0,    0, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

       0        // eod
};

void MyWebView::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    Q_UNUSED(_o);
    Q_UNUSED(_id);
    Q_UNUSED(_c);
    Q_UNUSED(_a);
}

const QMetaObject MyWebView::staticMetaObject = {
    { &QWebView::staticMetaObject, qt_meta_stringdata_MyWebView.data,
      qt_meta_data_MyWebView,  qt_static_metacall, nullptr, nullptr}
};


const QMetaObject *MyWebView::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *MyWebView::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_MyWebView.stringdata0))
        return static_cast<void*>(this);
    return QWebView::qt_metacast(_clname);
}

int MyWebView::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QWebView::qt_metacall(_c, _id, _a);
    return _id;
}
struct qt_meta_stringdata_CWebView_t {
    QByteArrayData data[27];
    char stringdata0[327];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_CWebView_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_CWebView_t qt_meta_stringdata_CWebView = {
    {
QT_MOC_LITERAL(0, 0, 8), // "CWebView"
QT_MOC_LITERAL(1, 9, 11), // "iconChanged"
QT_MOC_LITERAL(2, 21, 0), // ""
QT_MOC_LITERAL(3, 22, 12), // "loadFinished"
QT_MOC_LITERAL(4, 35, 2), // "ok"
QT_MOC_LITERAL(5, 38, 12), // "loadProgress"
QT_MOC_LITERAL(6, 51, 8), // "progress"
QT_MOC_LITERAL(7, 60, 11), // "loadStarted"
QT_MOC_LITERAL(8, 72, 16), // "selectionChanged"
QT_MOC_LITERAL(9, 89, 16), // "statusBarMessage"
QT_MOC_LITERAL(10, 106, 4), // "text"
QT_MOC_LITERAL(11, 111, 12), // "titleChanged"
QT_MOC_LITERAL(12, 124, 5), // "title"
QT_MOC_LITERAL(13, 130, 11), // "linkHovered"
QT_MOC_LITERAL(14, 142, 4), // "link"
QT_MOC_LITERAL(15, 147, 11), // "textContent"
QT_MOC_LITERAL(16, 159, 12), // "frameCreated"
QT_MOC_LITERAL(17, 172, 10), // "QWebFrame*"
QT_MOC_LITERAL(18, 183, 22), // "authenticationRequired"
QT_MOC_LITERAL(19, 206, 14), // "QNetworkReply*"
QT_MOC_LITERAL(20, 221, 5), // "reply"
QT_MOC_LITERAL(21, 227, 15), // "QAuthenticator*"
QT_MOC_LITERAL(22, 243, 13), // "authenticator"
QT_MOC_LITERAL(23, 257, 10), // "urlChanged"
QT_MOC_LITERAL(24, 268, 17), // "downloadRequested"
QT_MOC_LITERAL(25, 286, 15), // "QNetworkRequest"
QT_MOC_LITERAL(26, 302, 24) // "handleUnsupportedContent"

    },
    "CWebView\0iconChanged\0\0loadFinished\0"
    "ok\0loadProgress\0progress\0loadStarted\0"
    "selectionChanged\0statusBarMessage\0"
    "text\0titleChanged\0title\0linkHovered\0"
    "link\0textContent\0frameCreated\0QWebFrame*\0"
    "authenticationRequired\0QNetworkReply*\0"
    "reply\0QAuthenticator*\0authenticator\0"
    "urlChanged\0downloadRequested\0"
    "QNetworkRequest\0handleUnsupportedContent"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_CWebView[] = {

 // content:
       7,       // revision
       0,       // classname
       0,    0, // classinfo
      13,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

 // slots: name, argc, parameters, tag, flags
       1,    0,   79,    2, 0x0a /* Public */,
       3,    1,   80,    2, 0x0a /* Public */,
       5,    1,   83,    2, 0x0a /* Public */,
       7,    0,   86,    2, 0x0a /* Public */,
       8,    0,   87,    2, 0x0a /* Public */,
       9,    1,   88,    2, 0x0a /* Public */,
      11,    1,   91,    2, 0x0a /* Public */,
      13,    3,   94,    2, 0x0a /* Public */,
      16,    1,  101,    2, 0x0a /* Public */,
      18,    2,  104,    2, 0x0a /* Public */,
      23,    1,  109,    2, 0x0a /* Public */,
      24,    1,  112,    2, 0x0a /* Public */,
      26,    1,  115,    2, 0x0a /* Public */,

 // slots: parameters
    QMetaType::Void,
    QMetaType::Void, QMetaType::Bool,    4,
    QMetaType::Void, QMetaType::Int,    6,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::QString,   10,
    QMetaType::Void, QMetaType::QString,   12,
    QMetaType::Void, QMetaType::QString, QMetaType::QString, QMetaType::QString,   14,   12,   15,
    QMetaType::Void, 0x80000000 | 17,    2,
    QMetaType::Void, 0x80000000 | 19, 0x80000000 | 21,   20,   22,
    QMetaType::Void, QMetaType::QUrl,    2,
    QMetaType::Void, 0x80000000 | 25,    2,
    QMetaType::Void, 0x80000000 | 19,    2,

       0        // eod
};

void CWebView::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        CWebView *_t = static_cast<CWebView *>(_o);
        Q_UNUSED(_t)
        switch (_id) {
        case 0: _t->iconChanged(); break;
        case 1: _t->loadFinished((*reinterpret_cast< bool(*)>(_a[1]))); break;
        case 2: _t->loadProgress((*reinterpret_cast< int(*)>(_a[1]))); break;
        case 3: _t->loadStarted(); break;
        case 4: _t->selectionChanged(); break;
        case 5: _t->statusBarMessage((*reinterpret_cast< const QString(*)>(_a[1]))); break;
        case 6: _t->titleChanged((*reinterpret_cast< const QString(*)>(_a[1]))); break;
        case 7: _t->linkHovered((*reinterpret_cast< const QString(*)>(_a[1])),(*reinterpret_cast< const QString(*)>(_a[2])),(*reinterpret_cast< const QString(*)>(_a[3]))); break;
        case 8: _t->frameCreated((*reinterpret_cast< QWebFrame*(*)>(_a[1]))); break;
        case 9: _t->authenticationRequired((*reinterpret_cast< QNetworkReply*(*)>(_a[1])),(*reinterpret_cast< QAuthenticator*(*)>(_a[2]))); break;
        case 10: _t->urlChanged((*reinterpret_cast< const QUrl(*)>(_a[1]))); break;
        case 11: _t->downloadRequested((*reinterpret_cast< const QNetworkRequest(*)>(_a[1]))); break;
        case 12: _t->handleUnsupportedContent((*reinterpret_cast< QNetworkReply*(*)>(_a[1]))); break;
        default: ;
        }
    }
}

const QMetaObject CWebView::staticMetaObject = {
    { &QObject::staticMetaObject, qt_meta_stringdata_CWebView.data,
      qt_meta_data_CWebView,  qt_static_metacall, nullptr, nullptr}
};


const QMetaObject *CWebView::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *CWebView::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_CWebView.stringdata0))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int CWebView::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 13)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 13;
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 13)
            *reinterpret_cast<int*>(_a[0]) = -1;
        _id -= 13;
    }
    return _id;
}
QT_WARNING_POP
QT_END_MOC_NAMESPACE
